# Sink Custom Python Title 3 API library

[![PyPI version](https://img.shields.io/pypi/v/sink-pypi.svg)](https://pypi.org/project/sink-pypi/)

The Sink Custom Python Title 3 library provides convenient access to the Sink REST API from any Python 3.7+
application. The library includes type definitions for all request params and response fields,
and offers both synchronous and asynchronous clients powered by [httpx](https://github.com/encode/httpx).

## Documentation

The API documentation can be found [here](https://stainlessapi.com).

## Installation

```sh
pip install sink-pypi
```

## Usage

The full API of this library can be found in [api.md](https://www.github.com/stainless-sdks/sink-python-public/blob/master/api.md).

```python
import os
from sink import Sink

client = Sink(
    # This is the default and can be omitted
    user_token=os.environ.get("SINK_CUSTOM_API_KEY_ENV"),
    # defaults to "production".
    environment="sandbox",
    username="Robert",
    some_number_arg_required_no_default=0,
    some_number_arg_required_no_default_no_env=0,
    required_arg_no_env="<example>",
)

card = client.cards.create(
    type="SINGLE_USE",
    not_="TEST",
)
print(card.token)
```

While you can provide a `user_token` keyword argument,
we recommend using [python-dotenv](https://pypi.org/project/python-dotenv/)
to add `SINK_CUSTOM_API_KEY_ENV="My User Token"` to your `.env` file
so that your User Token is not stored in source control.

## Async usage

Simply import `AsyncSink` instead of `Sink` and use `await` with each API call:

```python
import os
import asyncio
from sink import AsyncSink

client = AsyncSink(
    # This is the default and can be omitted
    user_token=os.environ.get("SINK_CUSTOM_API_KEY_ENV"),
    # defaults to "production".
    environment="sandbox",
    username="Robert",
    some_number_arg_required_no_default=0,
    some_number_arg_required_no_default_no_env=0,
    required_arg_no_env="<example>",
)


async def main() -> None:
    card = await client.cards.create(
        type="SINGLE_USE",
        not_="TEST",
    )
    print(card.token)


asyncio.run(main())
```

Functionality between the synchronous and asynchronous clients is otherwise identical.

## Streaming Responses

We provide support for streaming responses using Server Side Events (SSE).

```python
from sink import Sink

client = Sink(
    username="Robert",
    some_number_arg_required_no_default=0,
    some_number_arg_required_no_default_no_env=0,
    required_arg_no_env="<example>",
)

stream = client.streaming.basic(
    model="string",
    prompt="string",
    stream=True,
)
for completion in stream:
    print(completion.completion)
```

The async client uses the exact same interface.

```python
from sink import Sink

client = Sink(
    username="Robert",
    some_number_arg_required_no_default=0,
    some_number_arg_required_no_default_no_env=0,
    required_arg_no_env="<example>",
)

stream = await client.streaming.basic(
    model="string",
    prompt="string",
    stream=True,
)
async for completion in stream:
    print(completion.completion)
```

## Using types

Nested request parameters are [TypedDicts](https://docs.python.org/3/library/typing.html#typing.TypedDict). Responses are [Pydantic models](https://docs.pydantic.dev), which provide helper methods for things like:

- Serializing back into JSON, `model.model_dump_json(indent=2, exclude_unset=True)`
- Converting to a dictionary, `model.model_dump(exclude_unset=True)`

Typed requests and responses provide autocomplete and documentation within your editor. If you would like to see type errors in VS Code to help catch bugs earlier, set `python.analysis.typeCheckingMode` to `basic`.

## Pagination

List methods in the Sink API are paginated.

This library provides auto-paginating iterators with each list response, so you do not have to request successive pages manually:

```python
import sink

client = Sink(
    username="Robert",
    some_number_arg_required_no_default=0,
    some_number_arg_required_no_default_no_env=0,
    required_arg_no_env="<example>",
)

all_cursors = []
# Automatically fetches more pages as needed.
for cursor in client.pagination_tests.cursor.list():
    # Do something with cursor here
    all_cursors.append(cursor)
print(all_cursors)
```

Or, asynchronously:

```python
import asyncio
import sink

client = AsyncSink(
    username="Robert",
    some_number_arg_required_no_default=0,
    some_number_arg_required_no_default_no_env=0,
    required_arg_no_env="<example>",
)


async def main() -> None:
    all_cursors = []
    # Iterate through items across all pages, issuing requests as needed.
    async for cursor in client.pagination_tests.cursor.list():
        all_cursors.append(cursor)
    print(all_cursors)


asyncio.run(main())
```

Alternatively, you can use the `.has_next_page()`, `.next_page_info()`, or `.get_next_page()` methods for more granular control working with pages:

```python
first_page = await client.pagination_tests.cursor.list()
if first_page.has_next_page():
    print(f"will fetch next page using these details: {first_page.next_page_info()}")
    next_page = await first_page.get_next_page()
    print(f"number of items we just fetched: {len(next_page.data)}")

# Remove `await` for non-async usage.
```

Or just work directly with the returned data:

```python
first_page = await client.pagination_tests.cursor.list()

print(f"next page cursor: {first_page.cursor}")  # => "next page cursor: ..."
for cursor in first_page.data:
    print(cursor.bar)

# Remove `await` for non-async usage.
```

## Nested params

Nested parameters are dictionaries, typed using `TypedDict`, for example:

```python
from sink import Sink

client = Sink(
    username="Robert",
    some_number_arg_required_no_default=0,
    some_number_arg_required_no_default_no_env=0,
    required_arg_no_env="<example>",
)

card = client.cards.create(
    type="SINGLE_USE",
)
print(card.token)
```

## File Uploads

Request parameters that correspond to file uploads can be passed as `bytes`, a [`PathLike`](https://docs.python.org/3/library/os.html#os.PathLike) instance or a tuple of `(filename, contents, media type)`.

```python
from pathlib import Path
from sink import Sink

client = Sink(
    username="Robert",
    some_number_arg_required_no_default=0,
    some_number_arg_required_no_default_no_env=0,
    required_arg_no_env="<example>",
)

client.files.create_multipart(
    file=Path("foo/bar.txt"),
    purpose="string",
)
```

The async client uses the exact same interface. If you pass a [`PathLike`](https://docs.python.org/3/library/os.html#os.PathLike) instance, the file contents will be read asynchronously automatically.

## Handling errors

When the library is unable to connect to the API (for example, due to network connection problems or a timeout), a subclass of `sink.APIConnectionError` is raised.

When the API returns a non-success status code (that is, 4xx or 5xx
response), a subclass of `sink.APIStatusError` is raised, containing `status_code` and `response` properties.

All errors inherit from `sink.APIError`.

```python
import sink
from sink import Sink

client = Sink(
    username="Robert",
    some_number_arg_required_no_default=0,
    some_number_arg_required_no_default_no_env=0,
    required_arg_no_env="<example>",
)

try:
    client.cards.create(
        type="an_incorrect_type",
    )
except sink.APIConnectionError as e:
    print("The server could not be reached")
    print(e.__cause__)  # an underlying Exception, likely raised within httpx.
except sink.RateLimitError as e:
    print("A 429 status code was received; we should back off a bit.")
except sink.APIStatusError as e:
    print("Another non-200-range status code was received")
    print(e.status_code)
    print(e.response)
```

Error codes are as followed:

| Status Code | Error Type                 |
| ----------- | -------------------------- |
| 400         | `BadRequestError`          |
| 401         | `AuthenticationError`      |
| 403         | `PermissionDeniedError`    |
| 404         | `NotFoundError`            |
| 422         | `UnprocessableEntityError` |
| 429         | `RateLimitError`           |
| >=500       | `InternalServerError`      |
| N/A         | `APIConnectionError`       |

### Retries

Certain errors are automatically retried 1 times by default, with a short exponential backoff.
Connection errors (for example, due to a network connectivity problem), 408 Request Timeout, 409 Conflict,
429 Rate Limit, and >=500 Internal errors are all retried by default.

You can use the `max_retries` option to configure or disable retry settings:

```python
from sink import Sink

# Configure the default for all requests:
client = Sink(
    # default is 2
    max_retries=0,
    username="Robert",
    some_number_arg_required_no_default=0,
    some_number_arg_required_no_default_no_env=0,
    required_arg_no_env="<example>",
)

# Or, configure per-request:
client.with_options(max_retries=5).cards.provision_foo(
    "my card token",
    digital_wallet="GOOGLE_PAY",
)
```

### Timeouts

By default requests time out after 1 minute. You can configure this with a `timeout` option,
which accepts a float or an [`httpx.Timeout`](https://www.python-httpx.org/advanced/#fine-tuning-the-configuration) object:

```python
from sink import Sink

# Configure the default for all requests:
client = Sink(
    # 20 seconds (default is 1 minute)
    timeout=20.0,
    username="Robert",
    some_number_arg_required_no_default=0,
    some_number_arg_required_no_default_no_env=0,
    required_arg_no_env="<example>",
)

# More granular control:
client = Sink(
    timeout=httpx.Timeout(60.0, read=5.0, write=10.0, connect=2.0),
    username="Robert",
    some_number_arg_required_no_default=0,
    some_number_arg_required_no_default_no_env=0,
    required_arg_no_env="<example>",
)

# Override per-request:
client.with_options(timeout=5 * 1000).cards.create(
    type="DIGITAL",
)
```

On timeout, an `APITimeoutError` is thrown.

Note that requests that time out are [retried twice by default](#retries).

## Default Headers

We automatically send the following headers with all requests.

| Header             | Value |
| ------------------ | ----- |
| `My-Api-Version`   | `11`  |
| `X-Enable-Metrics` | `1`   |

If you need to, you can override these headers by setting default headers per-request or on the client object.

```python
from sink import Sink

client = Sink(
    default_headers={"My-Api-Version": "My-Custom-Value"},
    username="Robert",
    some_number_arg_required_no_default=0,
    some_number_arg_required_no_default_no_env=0,
    required_arg_no_env="<example>",
)
```

## Advanced

### Logging

We use the standard library [`logging`](https://docs.python.org/3/library/logging.html) module.

You can enable logging by setting the environment variable `SINK_LOG` to `debug`.

```shell
$ export SINK_LOG=debug
```

### How to tell whether `None` means `null` or missing

In an API response, a field may be explicitly `null`, or missing entirely; in either case, its value is `None` in this library. You can differentiate the two cases with `.model_fields_set`:

```py
if response.my_field is None:
  if 'my_field' not in response.model_fields_set:
    print('Got json like {}, without a "my_field" key present at all.')
  else:
    print('Got json like {"my_field": null}.')
```

### Accessing raw response data (e.g. headers)

The "raw" Response object can be accessed by prefixing `.with_raw_response.` to any HTTP method call.

```py
from sink import Sink

client = Sink(
    username="Robert",
    some_number_arg_required_no_default=0,
    some_number_arg_required_no_default_no_env=0,
    required_arg_no_env="<example>",
)
response = client.cards.with_raw_response.create(
    type="SINGLE_USE",
    not_="TEST",
)
print(response.headers.get('X-My-Header'))

card = response.parse()  # get the object that `cards.create()` would have returned
print(card.token)
```

These methods return an [`APIResponse`](https://github.com/stainless-sdks/sink-python-public/tree/master/src/sink/_response.py) object.

### Configuring the HTTP client

You can directly override the [httpx client](https://www.python-httpx.org/api/#client) to customize it for your use case, including:

- Support for proxies
- Custom transports
- Additional [advanced](https://www.python-httpx.org/advanced/#client-instances) functionality

```python
import httpx
from sink import Sink

client = Sink(
    # Or use the `SINK_BASE_URL` env var
    base_url="http://my.test.server.example.com:8083",
    http_client=httpx.Client(
        proxies="http://my.test.proxy.example.com",
        transport=httpx.HTTPTransport(local_address="0.0.0.0"),
    ),
    username="Robert",
    some_number_arg_required_no_default=0,
    some_number_arg_required_no_default_no_env=0,
    required_arg_no_env="<example>",
)
```

### Managing HTTP resources

By default the library closes underlying HTTP connections whenever the client is [garbage collected](https://docs.python.org/3/reference/datamodel.html#object.__del__). You can manually close the client using the `.close()` method if desired, or with a context manager that closes when exiting.

## Versioning

This package generally follows [SemVer](https://semver.org/spec/v2.0.0.html) conventions, though certain backwards-incompatible changes may be released as minor versions:

1. Changes that only affect static types, without breaking runtime behavior.
2. Changes to library internals which are technically public but not intended or documented for external use. _(Please open a GitHub issue to let us know if you are relying on such internals)_.
3. Changes that we do not expect to impact the vast majority of users in practice.

We take backwards-compatibility seriously and work hard to ensure you can rely on a smooth upgrade experience.

We are keen for your feedback; please open an [issue](https://www.github.com/stainless-sdks/sink-python-public/issues) with questions, bugs, or suggestions.

## Requirements

Python 3.7 or higher.
